@@ -524,6 +524,13 @@ static bool match_interface_instance(const void *a, const void *b)
 	return false;
 }
 
+static bool match_interface_instance_ptr(const void *a, const void *b)
+{
+	const struct interface_instance *instance = a;
+
+	return instance->interface == b;
+}
+
 static void interface_add_record_free(void *data)
 {
 	struct interface_add_record *rec = data;
@@ -649,7 +656,8 @@ static struct object_node *makepath_recurse(struct object_node *node,
 	child = node->children;
 
 	while (child) {
-		if (!strncmp(child->subpath, path, end - path))
+		if (!strncmp(child->subpath, path, end - path) &&
+				child->subpath[end - path] == '\0')
 			goto done;
 
 		child = child->next;
@@ -690,7 +698,8 @@ static struct object_node *lookup_recurse(struct object_node *node,
 	child = node->children;
 
 	while (child) {
-		if (!strncmp(child->subpath, path, end - path))
+		if (!strncmp(child->subpath, path, end - path) &&
+				child->subpath[end - path] == '\0')
 			return lookup_recurse(child->node, end);
 
 		child = child->next;
@@ -1221,10 +1230,12 @@ bool _dbus_object_tree_property_changed(struct l_dbus *dbus,
 	return true;
 }
 
-static void set_property_complete(struct l_dbus *dbus,
+static void pending_property_set_done(struct l_dbus *dbus,
 					struct l_dbus_message *message,
 					struct l_dbus_message *reply)
 {
+	const char *member;
+	const char *interface_name;
 	const char *property_name;
 	struct l_dbus_message_iter variant;
 
@@ -1233,19 +1244,28 @@ static void set_property_complete(struct l_dbus *dbus,
 		l_dbus_message_set_arguments(reply, "");
 	}
 
-	l_dbus_send(dbus, reply);
+	l_dbus_send(dbus, l_dbus_message_ref(reply));
 
-	if (!l_dbus_message_is_error(reply)) {
-		l_dbus_message_get_arguments(message, "sv", &property_name,
-						&variant);
+	member = l_dbus_message_get_member(message);
+	if (!strcmp(member, "SetProperty")) {
+		if (!l_dbus_message_get_arguments(message, "sv",
+						&property_name, &variant))
+			goto done;
 
-		_dbus_object_tree_property_changed(dbus,
+		interface_name = l_dbus_message_get_interface(message);
+	} else if (strcmp(member, "Set") ||
+			!l_dbus_message_get_arguments(message, "ssv",
+							&interface_name,
+							&property_name,
+							&variant))
+		goto done;
+
+	_dbus_object_tree_property_changed(dbus,
 					l_dbus_message_get_path(message),
-					l_dbus_message_get_interface(message),
-					property_name);
-	}
-
+					interface_name, property_name);
+done:
 	l_dbus_message_unref(message);
+	l_dbus_message_unref(reply);
 }
 
 static struct l_dbus_message *old_set_property(struct l_dbus *dbus,
@@ -1286,10 +1306,10 @@ static struct l_dbus_message *old_set_property(struct l_dbus *dbus,
 						property_name);
 
 	reply = property->setter(dbus, l_dbus_message_ref(message), &variant,
-					set_property_complete, user_data);
+					pending_property_set_done, user_data);
 
 	if (reply)
-		set_property_complete(dbus, message, reply);
+		pending_property_set_done(dbus, message, reply);
 
 	return NULL;
 }
@@ -1407,6 +1427,43 @@ bool _dbus_object_tree_unregister_interface(struct _dbus_object_tree *tree,
 	return true;
 }
 
+static void collect_instances(struct object_node *node,
+				const char *path,
+				struct l_queue *announce)
+{
+	const struct l_queue_entry *entry;
+	struct interface_add_record *change_rec;
+	const struct child_node *child;
+
+	if (!node->instances)
+		goto recurse;
+
+	change_rec = l_new(struct interface_add_record, 1);
+	change_rec->path = l_strdup(path);
+	change_rec->object = node;
+	change_rec->instances = l_queue_new();
+
+	for (entry = l_queue_get_entries(node->instances); entry;
+			entry = entry->next)
+		l_queue_push_tail(change_rec->instances, entry->data);
+
+	l_queue_push_tail(announce, change_rec);
+
+recurse:
+	if (!strcmp(path, "/"))
+		path = "";
+
+	for (child = node->children; child; child = child->next) {
+		char *child_path;
+
+		child_path = l_strdup_printf("%s/%s", path, child->subpath);
+
+		collect_instances(child->node, child_path, announce);
+
+		l_free(child_path);
+	}
+}
+
 static bool match_interfaces_added_object(const void *a, const void *b)
 {
 	const struct interface_add_record *rec = a;
@@ -1449,8 +1506,7 @@ bool _dbus_object_tree_add_interface(struct _dbus_object_tree *tree,
 	 * Check to make sure we do not have this interface already
 	 * registered for this object
 	 */
-	if (l_queue_find(object->instances, match_interface_instance,
-				(char *) interface))
+	if (l_queue_find(object->instances, match_interface_instance_ptr, dbi))
 		return false;
 
 	instance = l_new(struct interface_instance, 1);
@@ -1464,7 +1520,7 @@ bool _dbus_object_tree_add_interface(struct _dbus_object_tree *tree,
 		manager = entry->data;
 		path_len = strlen(manager->path);
 
-		if (memcmp(path, manager->path, path_len) ||
+		if (strncmp(path, manager->path, path_len) ||
 				(path[path_len] != '\0' &&
 				 path[path_len] != '/' && path_len > 1))
 			continue;
@@ -1495,11 +1551,36 @@ bool _dbus_object_tree_add_interface(struct _dbus_object_tree *tree,
 		manager->announce_removed = l_queue_new();
 
 		l_queue_push_tail(tree->object_managers, manager);
+
+		/* Emit InterfacesAdded for interfaces added before OM */
+		collect_instances(object, path, manager->announce_added);
+
+		if (manager->dbus && !l_queue_isempty(manager->announce_added))
+			schedule_emit_signals(manager->dbus);
 	}
 
 	return true;
 }
 
+void *_dbus_object_tree_get_interface_data(struct _dbus_object_tree *tree,
+						const char *path,
+						const char *interface)
+{
+	struct object_node *object;
+	struct interface_instance *instance;
+
+	object = l_hashmap_lookup(tree->objects, path);
+	if (!object)
+		return NULL;
+
+	instance = l_queue_find(object->instances, match_interface_instance,
+				(char *) interface);
+	if (!instance)
+		return NULL;
+
+	return instance->user_data;
+}
+
 static bool match_object_manager_path(const void *a, const void *b)
 {
 	const struct object_manager *manager = a;
@@ -1528,14 +1609,13 @@ bool _dbus_object_tree_remove_interface(struct _dbus_object_tree *tree,
 	if (!instance)
 		return false;
 
-	interface_instance_free(instance);
-
 	if (!strcmp(interface, L_DBUS_INTERFACE_OBJECT_MANAGER)) {
 		manager = l_queue_remove_if(tree->object_managers,
 						match_object_manager_path,
 						(char *) path);
 
-		object_manager_free(manager);
+		if (manager)
+			object_manager_free(manager);
 	}
 
 	for (entry = l_queue_get_entries(tree->object_managers); entry;
@@ -1543,7 +1623,7 @@ bool _dbus_object_tree_remove_interface(struct _dbus_object_tree *tree,
 		manager = entry->data;
 		path_len = strlen(manager->path);
 
-		if (memcmp(path, manager->path, path_len) ||
+		if (strncmp(path, manager->path, path_len) ||
 				(path[path_len] != '\0' &&
 				 path[path_len] != '/' && path_len > 1))
 			continue;
@@ -1590,6 +1670,8 @@ bool _dbus_object_tree_remove_interface(struct _dbus_object_tree *tree,
 	if (property_change_rec)
 		property_change_record_free(property_change_rec);
 
+	interface_instance_free(instance);
+
 	return true;
 }
 
@@ -1769,32 +1851,6 @@ static struct l_dbus_message *properties_get(struct l_dbus *dbus,
 	return reply;
 }
 
-static void properties_set_complete(struct l_dbus *dbus,
-					struct l_dbus_message *message,
-					struct l_dbus_message *reply)
-{
-	const char *interface_name, *property_name;
-	struct l_dbus_message_iter variant;
-
-	if (!reply) {
-		reply = l_dbus_message_new_method_return(message);
-		l_dbus_message_set_arguments(reply, "");
-	}
-
-	l_dbus_send(dbus, reply);
-
-	if (!l_dbus_message_is_error(reply)) {
-		l_dbus_message_get_arguments(message, "ssv", &interface_name,
-						&property_name, &variant);
-
-		_dbus_object_tree_property_changed(dbus,
-					l_dbus_message_get_path(message),
-					interface_name, property_name);
-	}
-
-	l_dbus_message_unref(message);
-}
-
 static struct l_dbus_message *properties_set(struct l_dbus *dbus,
 						struct l_dbus_message *message,
 						void *user_data)
@@ -1843,8 +1899,7 @@ static struct l_dbus_message *properties_set(struct l_dbus *dbus,
 	/* If we got here the object must exist */
 
 	instance = l_queue_find(object->instances,
-				match_interface_instance,
-				(char *) interface_name);
+				match_interface_instance_ptr, interface);
 	if (!instance)
 		return l_dbus_message_new_error(message,
 						"org.freedesktop.DBus.Error."
@@ -1853,11 +1908,11 @@ static struct l_dbus_message *properties_set(struct l_dbus *dbus,
 						interface_name);
 
 	reply = property->setter(dbus, l_dbus_message_ref(message), &variant,
-					properties_set_complete,
+					pending_property_set_done,
 					instance->user_data);
 
 	if (reply)
-		properties_set_complete(dbus, message, reply);
+		pending_property_set_done(dbus, message, reply);
 
 	return NULL;
 }
@@ -1986,17 +2041,16 @@ recurse:
 	return true;
 }
 
-static struct l_dbus_message *get_managed_objects(struct l_dbus *dbus,
-						struct l_dbus_message *message,
-						void *user_data)
+struct l_dbus_message *_dbus_object_tree_get_objects(
+						struct _dbus_object_tree *tree,
+						struct l_dbus *dbus,
+						const char *path,
+						struct l_dbus_message *message)
 {
-	struct _dbus_object_tree *tree = _dbus_get_tree(dbus);
 	const struct object_node *node;
 	struct l_dbus_message *reply;
 	struct l_dbus_message_builder *builder;
-	const char *path;
 
-	path = l_dbus_message_get_path(message);
 	node = l_hashmap_lookup(tree->objects, path);
 
 	reply = l_dbus_message_new_method_return(message);
@@ -2023,6 +2077,16 @@ static struct l_dbus_message *get_managed_objects(struct l_dbus *dbus,
 	return reply;
 }
 
+static struct l_dbus_message *get_managed_objects(struct l_dbus *dbus,
+						struct l_dbus_message *message,
+						void *user_data)
+{
+	struct _dbus_object_tree *tree = _dbus_get_tree(dbus);
+	const char *path = l_dbus_message_get_path(message);
+
+	return _dbus_object_tree_get_objects(tree, dbus, path, message);
+}
+
 static void object_manager_setup_func(struct l_dbus_interface *interface)
 {
 	l_dbus_interface_method(interface, "GetManagedObjects", 0,
