@@ -78,9 +78,11 @@ uint16_t _dhcp_checksumv(const struct iovec *iov, size_t iov_cnt)
 			sum += check[i];
 	}
 
+	j--;
+
 	if (len & 0x01) {
 		const uint8_t *odd = iov[j].iov_base;
-		sum += odd[len - 1];
+		sum += odd[iov[j].iov_len - 1];
 	}
 
 	while (sum >> 16)
@@ -303,7 +305,7 @@ static int kernel_udp_socket_open(const char *ifname,
 	return s;
 
 error:
-	TEMP_FAILURE_RETRY(close(s));
+	L_TFR(close(s));
 	return -errno;
 }
 
@@ -442,7 +444,7 @@ static int kernel_raw_socket_open(uint32_t ifindex, uint16_t port, uint32_t xid)
 	return s;
 
 error:
-	TEMP_FAILURE_RETRY(close(s));
+	L_TFR(close(s));
 	return -errno;
 }
 
@@ -477,7 +479,7 @@ static void _dhcp_default_transport_close(struct dhcp_transport *s)
 	transport->io = NULL;
 
 	if (transport->udp_fd >= 0) {
-		TEMP_FAILURE_RETRY(close(transport->udp_fd));
+		L_TFR(close(transport->udp_fd));
 		transport->udp_fd = -1;
 	}
 }
